# Library imports
from arrow import Arrow

# Local imports
from . import validators

DUST = 546
PUBLIC_KEY_BYTES_LENGTH = 33
SCRIPT_INT_MAX_WHEN_JAVASCRIPT = int(2 ** 53)
SATS_PER_BCH = int(1e8)
MAX_REASONABLE_SATS = int(10e6 * SATS_PER_BCH)  # 10 million BCH
MIN_NLOCKTIME_TIMESTAMP = 500_000_000


class PublicKey(str):
    """Validated BCH public key that can be used as a string."""
    def __init__(self, value):
        super().__init__()

        # don't allow silent coercion
        validators.instance(value, str)

        # use a strict format
        if self[:2].lower() == '0x':
            raise ValueError('public key should not include the 0x prefix')

        # require hex value validity
        try:
            bytes.fromhex(self)
        except Exception as e:
            raise ValueError(f'unable to interpret {self} as a hex string: {e}')

        # require strict length
        if len(self) != 2 * PUBLIC_KEY_BYTES_LENGTH:
            raise ValueError(f'public key should be 33 bytes but got {len(self) / 2} ({self})')


class ScriptTimestamp(int):
    """Validated BCH timestamp (seconds) that can be used as an integer."""
    def __init__(self, value):
        super().__init__()
        validators.instance(value, int)  # i.e. don't allow silent coercion
        validators.less_equal(self, SCRIPT_INT_MAX_WHEN_JAVASCRIPT)
        validators.greater_equal(self, MIN_NLOCKTIME_TIMESTAMP)

    def __str__(self):
        return f'{int(self)} ({Arrow.utcfromtimestamp(self)})'

    def __repr__(self):
        return self.__str__()

    @classmethod
    def earliest_possible(cls):
        return cls(MIN_NLOCKTIME_TIMESTAMP)

    @classmethod
    def latest_possible(cls):
        return cls(SCRIPT_INT_MAX_WHEN_JAVASCRIPT)


class Sats(int):
    """Validated amount of sats that can be used in calculations (positive or negative)."""
    _min_value = -MAX_REASONABLE_SATS

    def __init__(self, value):
        super().__init__()
        validators.instance(value, int)  # i.e. don't allow silent coercion
        validators.less_equal(self, MAX_REASONABLE_SATS)
        validators.greater_equal(self, self._min_value)

    @property
    def bch(self) -> float:
        return round(self / SATS_PER_BCH, ndigits=8)


class UtxoSats(Sats):
    """Validated amount of sats that can exist in a utxo."""
    _min_value = DUST



# Built-in imports
import unittest

# Local imports
from .bch_primitives import Sats
from .role import Role
from .fee import (
    fees_from_role,
    fees_to_role,
    FeeAgreement,
)


fee_100000_taker_to_ss = FeeAgreement(name='a', amount_sats=Sats(100000), receiving=Role.SETTLEMENT_SERVICE, paying=Role.TAKER)
fee_200000_maker_to_ss = FeeAgreement(name='b', amount_sats=Sats(200000), receiving=Role.SETTLEMENT_SERVICE, paying=Role.MAKER)
fee_400000_ss_to_maker = FeeAgreement(name='c', amount_sats=Sats(400000), receiving=Role.MAKER, paying=Role.SETTLEMENT_SERVICE)
set_of_fees = [fee_100000_taker_to_ss, fee_200000_maker_to_ss, fee_400000_ss_to_maker]
expected_fees_from_ss = [fee_400000_ss_to_maker]
expected_fees_to_ss = [fee_100000_taker_to_ss, fee_200000_maker_to_ss]


class TestFeeFilters(unittest.TestCase):
    def test_fees_from_role(self):
        fees_from_taker = fees_from_role(fees=set_of_fees, role=Role.SETTLEMENT_SERVICE)
        self.assertCountEqual(fees_from_taker, expected_fees_from_ss)

    def test_fees_to_role(self):
        fees_to_ss = fees_to_role(fees=set_of_fees, role=Role.SETTLEMENT_SERVICE)
        self.assertCountEqual(fees_to_ss, expected_fees_to_ss)



# Built-in imports
import unittest

# Local imports
from . import bch_primitives as bchp


VALID_PUBLIC_KEY_HEX = '0123456789abcdef0123456789ABCDEF0123456789abcdef0123456789abcdef01'
VALID_TIMESTAMP_INT = 5_000_000_000
VALID_SATS = 1


class TestPublicKey(unittest.TestCase):
    def test_value_is_string_matching_source_public_key_hex(self):
        pk = bchp.PublicKey(VALID_PUBLIC_KEY_HEX)
        self.assertEqual(pk, VALID_PUBLIC_KEY_HEX)

    def test_ValueError_if_source_has_hex_prefix(self):
        with self.assertRaises(ValueError):
            valid_key_with_prefx = f'0x{VALID_PUBLIC_KEY_HEX}'
            bchp.PublicKey(valid_key_with_prefx)

    def test_ValueError_if_source_is_not_33_bytes_66_hex_chars(self):
        self.assertEqual(len(VALID_PUBLIC_KEY_HEX), 66)
        with self.assertRaises(ValueError):
            too_short = VALID_PUBLIC_KEY_HEX[2:]
            bchp.PublicKey(too_short)
        with self.assertRaises(ValueError):
            too_long = VALID_PUBLIC_KEY_HEX + '00'
            bchp.PublicKey(too_long)
        with self.assertRaises(ValueError):
            odd_length = VALID_PUBLIC_KEY_HEX + '0'
            bchp.PublicKey(odd_length)


class TestScriptTimestamp(unittest.TestCase):
    def test_value_is_integer_matching_source(self):
        timestamp = bchp.ScriptTimestamp(VALID_TIMESTAMP_INT)
        self.assertEqual(timestamp, VALID_TIMESTAMP_INT)

    def test_AssertionError_if_source_is_not_int(self):
        with self.assertRaises(AssertionError):
            bchp.ScriptTimestamp(VALID_TIMESTAMP_INT + 0.1)

    def test_AssertionError_if_source_greater_than_safe_javascript_integer_as_conservative_proxy_for_script_integers(self):
        basically_max_javascript_safe_int = 2**53
        # should work
        bchp.ScriptTimestamp(basically_max_javascript_safe_int)

        # should fail
        with self.assertRaises(AssertionError):
            bchp.ScriptTimestamp(basically_max_javascript_safe_int + 1)

    def test_AssertionError_if_source_less_than_minimum_valid_nlocktime_timestamp(self):
        # should work
        bchp.ScriptTimestamp(500_000_000)

        # should fail
        with self.assertRaises(AssertionError):
            bchp.ScriptTimestamp(500_000_000 - 1)


class TestSats(unittest.TestCase):
    def test_value_is_integer_matching_source(self):
        sats = bchp.Sats(VALID_SATS)
        self.assertEqual(sats, VALID_SATS)

    def test_AssertionError_if_source_is_not_int(self):
        with self.assertRaises(AssertionError):
            bchp.Sats(VALID_SATS + 0.1)

    def test_AssertionError_if_source_is_less_or_greater_than_plus_minus_arbitrary_max_reasonable_sats(self):
        # fails above max
        with self.assertRaises(AssertionError):
            bchp.Sats(bchp.MAX_REASONABLE_SATS + 1)

        # fails below negative max
        with self.assertRaises(AssertionError):
            bchp.Sats(-bchp.MAX_REASONABLE_SATS - 1)

    def test_returns_eight_digits_of_float_bch_value(self):
        original_sats = bchp.Sats(1234)
        bch_equivalent = original_sats.bch
        self.assertEqual(bch_equivalent, 0.00001234)


class TestUtxoSats(unittest.TestCase):
    def test_behaves_mostly_like_sats(self):
        self.assertTrue(issubclass(bchp.UtxoSats, bchp.Sats))

    def test_AssertionError_if_source_is_less_than_positive_dust(self):
        # works
        bchp.UtxoSats(bchp.DUST)

        # doesn't work
        with self.assertRaises(AssertionError):
            bchp.UtxoSats(bchp.DUST - 1)


if __name__ == '__main__':
    unittest.main()



# Built-in imports
import unittest

# Library imports
from arrow import Arrow

# Local imports
from .bch_primitives import (
    SATS_PER_BCH,
    PublicKey,
    Sats,
    ScriptTimestamp,
    UtxoSats,
)
from .contract import (
    ContractFunding,
    ContractProposal,
    LongLeverage,
    NominalOracleUnitsXSatsPerBch,
    RedemptionType,
    Side,
    UnredeemableError,
)
from .fee import FeeAgreement
from .oracle import (
    OracleUnit,
    ScriptPriceInOracleUnitsPerBch,
    UsdEM2Beta,
)
from .role import Role


###################
# Contract Tests
###################
# Generic intent values, not always perfectly reproducible from the contract
START_TIMESTAMP = ScriptTimestamp(Arrow(year=2021, month=10, day=21).int_timestamp)
MATURITY_TIMESTAMP = ScriptTimestamp(Arrow(year=2021, month=10, day=23).int_timestamp)  # 2 days later
NOMINAL_USDEM2 = UsdEM2Beta(100_00.0)  # $100.00
LONG_LEVERAGE = LongLeverage(4.5)
START_PRICE_USDEM2_PER_BCH = ScriptPriceInOracleUnitsPerBch(200_00)  # $200.00 / BCH

# Derived values

# The compound nominal value is a giant number stored in the original AnyHedge contract and needed
# to get around the lack of multiplication by optimizing the contract into a single division operation
# = NominalOracleUnitsXSatsPerBch(round(NOMINAL_USDEM2 * SATS_PER_BCH))
DERIVED_NOMINAL_UNITS_X_SATS_PER_BCH = NominalOracleUnitsXSatsPerBch(1_000_000_000_000)

# Liquidation Price is the price where total bch covers exactly nominal units
# Calculate as round(START_PRICE_USDEM2_PER_BCH * (1 - 1 / LONG_LEVERAGE))
# $155.56 / BCH
DERIVED_LOW_LIQUIDATION_PRICE = ScriptPriceInOracleUnitsPerBch(155_56)

# The bch amount to cover the initial nominal units at the start price
# = round(NOMINAL_USDEM2 * SATS_PER_BCH / START_PRICE_USDEM2_PER_BCH)
# 0.5 BCH
DERIVED_HEDGE_INPUT_SATS = UtxoSats(50_000_000)

# The total bch amount needed to cover nominal units at liquidation
# Calculated as ceil((NOMINAL_USDEM2 * SATS_PER_BCH) / DERIVED_LIQUIDATION_PRICE_UNITS_PER_BCH)
DERIVED_TOTAL_INPUT_SATS = UtxoSats(64_283_878)

# The bch amount to provide volatility protection as specified by the multiplier
# This is the amount to cover the difference between hedge bch at start and hedge bch at liquidation
# Calculated as DERIVED_TOTAL_INPUT_SATS - DERIVED_HEDGE_INPUT_SATS
DERIVED_LONG_INPUT_SATS = UtxoSats(14_283_878)

# Converting the input amounts into original units at start
# for hedge, it's DERIVED_NOMINAL_UNITS_X_SATS_PER_BCH / SATS_PER_BCH
DERIVED_HEDGE_INPUT_ORACLE_UNITS = UsdEM2Beta(100_00.0)
# for long, it's DERIVED_LONG_INPUT_SATS * START_PRICE_USDEM2_PER_BCH / SATS_PER_BCH
DERIVED_LONG_INPUT_ORACLE_UNITS = UsdEM2Beta(2856.7756)


def standard_contract_proposal(
    start_timestamp: ScriptTimestamp = START_TIMESTAMP,
    maturity_timestamp: ScriptTimestamp = MATURITY_TIMESTAMP,
    nominal_oracleUnits: OracleUnit = NOMINAL_USDEM2,
    long_leverage: LongLeverage = LONG_LEVERAGE,
    start_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch = START_PRICE_USDEM2_PER_BCH,
    oracle_public_key: PublicKey = NOMINAL_USDEM2.public_key,
    maker_side: Side = Side.HEDGE,
) -> ContractProposal:
    return ContractProposal.new_from_intent(
        start_timestamp=start_timestamp,
        maturity_timestamp=maturity_timestamp,
        nominal_oracleUnits=nominal_oracleUnits,
        long_leverage=long_leverage,
        start_price_oracleUnits_per_bch=start_price_oracleUnits_per_bch,
        maker_side=maker_side,
    )


class TestContractProposalDerivedValues(unittest.TestCase):
    ######################
    # Exact contract parameters, some root values, some derived from root values
    ######################
    def test_nominal_oracleUnits_x_satsPerBch(self):
        self.assertEqual(standard_contract_proposal().nominal_oracleUnits_x_satsPerBch, DERIVED_NOMINAL_UNITS_X_SATS_PER_BCH)

    def test_low_liquidation_price_oracleUnits_per_bch(self):
        self.assertEqual(standard_contract_proposal().low_liquidation_price_oracleUnits_per_bch, DERIVED_LOW_LIQUIDATION_PRICE)

    def test_maker_side(self):
        self.assertEqual(standard_contract_proposal().maker_side, Side.HEDGE)

    ######################
    # Exact secondary values derived from contract parameters
    ######################
    def test_hedge_input_sats(self):
        self.assertEqual(standard_contract_proposal().hedge_input_sats, DERIVED_HEDGE_INPUT_SATS)

    def test_total_input_sats(self):
        self.assertEqual(standard_contract_proposal().total_input_sats, DERIVED_TOTAL_INPUT_SATS)

    def test_long_input_sats(self):
        self.assertEqual(standard_contract_proposal().long_input_sats, DERIVED_LONG_INPUT_SATS)

    ######################
    # Reverse calculated intent values from the contract's perspective
    ######################
    def test_effective_nominal_oracleUnits(self):
        self.assertAlmostEqual(standard_contract_proposal().effective_nominal_oracleUnits, NOMINAL_USDEM2, delta=.001)

    def test_effective_long_leverage(self):
        self.assertAlmostEqual(standard_contract_proposal().effective_long_leverage, LONG_LEVERAGE, delta=.001)

    ######################
    # Sided versions of values
    ######################
    def test_all_sided_values(self):
        hedge_maker_contract = standard_contract_proposal(maker_side=Side.HEDGE)
        self.assertEqual(hedge_maker_contract.maker_side, Side.HEDGE)
        self.assertEqual(hedge_maker_contract.taker_side, Side.LONG)
        self.assertEqual(hedge_maker_contract.maker_input_sats, DERIVED_HEDGE_INPUT_SATS)
        self.assertEqual(hedge_maker_contract.taker_input_sats, DERIVED_LONG_INPUT_SATS)
        self.assertEqual(hedge_maker_contract.maker_input_oracleUnits, DERIVED_HEDGE_INPUT_ORACLE_UNITS)
        self.assertEqual(hedge_maker_contract.taker_input_oracleUnits, DERIVED_LONG_INPUT_ORACLE_UNITS)

        long_maker_contract = standard_contract_proposal(maker_side=Side.LONG)
        self.assertEqual(long_maker_contract.maker_side, Side.LONG)
        self.assertEqual(long_maker_contract.taker_side, Side.HEDGE)
        self.assertEqual(long_maker_contract.maker_input_sats, DERIVED_LONG_INPUT_SATS)
        self.assertEqual(long_maker_contract.taker_input_sats, DERIVED_HEDGE_INPUT_SATS)
        self.assertEqual(long_maker_contract.maker_input_oracleUnits, DERIVED_LONG_INPUT_ORACLE_UNITS)
        self.assertEqual(long_maker_contract.taker_input_oracleUnits, DERIVED_HEDGE_INPUT_ORACLE_UNITS)


class TestContractProposalValidation(unittest.TestCase):
    def test_ValueError_if_duration_too_short(self):
        with self.assertRaises(ValueError):
            too_short_maturity_timestamp = ScriptTimestamp(START_TIMESTAMP + 60 - 1)
            standard_contract_proposal(maturity_timestamp=too_short_maturity_timestamp)


# Create the standard set of fees
FEE_100000_TAKER_TO_MAKER = FeeAgreement(name='fee type a', amount_sats=Sats(100000), receiving=Role.MAKER, paying=Role.TAKER)
FEE_200000_TAKER_TO_MAKER = FeeAgreement(name='fee type b', amount_sats=Sats(200000), receiving=Role.MAKER, paying=Role.TAKER)
FEE_400000_TAKER_TO_SETTLEMENT_SERVICE = FeeAgreement(name='fee type ss', amount_sats=Sats(400000), receiving=Role.SETTLEMENT_SERVICE, paying=Role.TAKER)
FEES = [FEE_100000_TAKER_TO_MAKER, FEE_200000_TAKER_TO_MAKER, FEE_400000_TAKER_TO_SETTLEMENT_SERVICE]


# Group them for reference
TOTAL_FEES_TO_MAKER_SATS = Sats(
    + FEE_100000_TAKER_TO_MAKER.amount_sats
    + FEE_200000_TAKER_TO_MAKER.amount_sats
)
TOTAL_FEES_TO_TAKER_SATS = Sats(
    - FEE_100000_TAKER_TO_MAKER.amount_sats
    - FEE_200000_TAKER_TO_MAKER.amount_sats
    - FEE_400000_TAKER_TO_SETTLEMENT_SERVICE.amount_sats
)


def standard_contract_funding(proposal: ContractProposal | None = None) -> ContractFunding:
    proposal = proposal or standard_contract_proposal()
    return proposal.fund(fee_agreements=FEES)


class TestContractFundingDerivedValues(unittest.TestCase):
    def test_effective_adjusted_price_when_maker_is_hedge(self):
        # See the implementation for explanation of the formula
        # base values
        funding = standard_contract_funding()
        u = funding.base_proposal.effective_nominal_oracleUnits
        pS = float(funding.base_proposal.start_price_oracleUnits_per_bch)

        # in the context of long taker
        fP = -TOTAL_FEES_TO_TAKER_SATS.bch
        fR = 0.0
        expected_adjusted_price_for_taker = u * pS / (u - pS * (fP - fR))
        adjusted_price_for_taker = funding.effective_adjusted_price_oracleUnits_per_bch(Role.TAKER)
        self.assertAlmostEqual(adjusted_price_for_taker, expected_adjusted_price_for_taker)

        # in the context of hedge maker
        fP = 0.0
        fR = +TOTAL_FEES_TO_MAKER_SATS.bch
        expected_adjusted_price_for_maker = u * pS / (u - pS * (fP - fR))
        adjusted_price_for_maker = funding.effective_adjusted_price_oracleUnits_per_bch(Role.MAKER)
        self.assertAlmostEqual(adjusted_price_for_maker, expected_adjusted_price_for_maker)

        # Sanity check that maker and taker with different views on the fees
        # have different apparent adjusted prices
        self.assertNotEqual(expected_adjusted_price_for_taker, expected_adjusted_price_for_maker)

        # Sanity check that the result prices are close to the original price
        one_percent_of_original_price = 0.05 * pS
        self.assertAlmostEqual(adjusted_price_for_taker, pS, delta=one_percent_of_original_price)
        self.assertAlmostEqual(adjusted_price_for_maker, pS, delta=one_percent_of_original_price)

    def test_effective_adjusted_price_when_maker_is_long(self):
        # base values
        funding = standard_contract_funding(standard_contract_proposal(maker_side=Side.LONG))
        u = funding.base_proposal.effective_nominal_oracleUnits
        pS = float(funding.base_proposal.start_price_oracleUnits_per_bch)

        # in the context of hedge taker
        fP = +TOTAL_FEES_TO_TAKER_SATS.bch
        fR = 0.0
        expected_adjusted_price_for_taker = u * pS / (u - pS * (fP - fR))
        adjusted_price_for_taker = funding.effective_adjusted_price_oracleUnits_per_bch(Role.TAKER)
        self.assertAlmostEqual(adjusted_price_for_taker, expected_adjusted_price_for_taker)

        # in the context of long maker
        fP = 0.0
        fR = -TOTAL_FEES_TO_MAKER_SATS.bch
        expected_adjusted_price_for_maker = u * pS / (u - pS * (fP - fR))
        adjusted_price_for_maker = funding.effective_adjusted_price_oracleUnits_per_bch(Role.MAKER)
        self.assertAlmostEqual(adjusted_price_for_maker, expected_adjusted_price_for_maker)

        # Sanity check that maker and taker with different views on the fees
        # have different apparent adjusted prices
        self.assertNotEqual(expected_adjusted_price_for_taker, expected_adjusted_price_for_maker)

        # Sanity check that the result prices are close to the original price
        one_percent_of_original_price = 0.05 * pS
        self.assertAlmostEqual(adjusted_price_for_taker, pS, delta=one_percent_of_original_price)
        self.assertAlmostEqual(adjusted_price_for_maker, pS, delta=one_percent_of_original_price)

    def test_fee_sats_to_side(self):
        # Confirm hedge maker and long taker
        hedge_maker_contract = standard_contract_funding(standard_contract_proposal(maker_side=Side.HEDGE))
        long_maker_contract = standard_contract_funding(standard_contract_proposal(maker_side=Side.LONG))

        expected_fees_sats_to_maker = TOTAL_FEES_TO_MAKER_SATS
        expected_fees_sats_to_taker = TOTAL_FEES_TO_TAKER_SATS

        # Hedge Maker
        self.assertEqual(hedge_maker_contract.fee_sats_to_hedge, expected_fees_sats_to_maker)
        self.assertEqual(hedge_maker_contract.fee_sats_to_maker, expected_fees_sats_to_maker)

        # Long Taker
        self.assertEqual(hedge_maker_contract.fee_sats_to_long, expected_fees_sats_to_taker)
        self.assertEqual(hedge_maker_contract.fee_sats_to_taker, expected_fees_sats_to_taker)

        # Long Maker
        self.assertEqual(long_maker_contract.fee_sats_to_long, expected_fees_sats_to_maker)
        self.assertEqual(long_maker_contract.fee_sats_to_maker, expected_fees_sats_to_maker)

        # Hedge Taker
        self.assertEqual(long_maker_contract.fee_sats_to_hedge, expected_fees_sats_to_taker)
        self.assertEqual(long_maker_contract.fee_sats_to_taker, expected_fees_sats_to_taker)

    def test_fee_oracleUnits_to_side(self):
        # Confirm hedge maker and long taker
        hedge_maker_contract = standard_contract_funding(standard_contract_proposal(maker_side=Side.HEDGE))
        long_maker_contract = standard_contract_funding(standard_contract_proposal(maker_side=Side.LONG))
        oracle_class = hedge_maker_contract.base_proposal.oracle_unit_cls

        expected_fee_oracleUnits_to_maker_at_funding = oracle_class(TOTAL_FEES_TO_MAKER_SATS * START_PRICE_USDEM2_PER_BCH / SATS_PER_BCH)
        expected_fee_oracleUnits_to_taker_at_funding = oracle_class(TOTAL_FEES_TO_TAKER_SATS * START_PRICE_USDEM2_PER_BCH / SATS_PER_BCH)

        # Hedge Maker
        self.assertEqual(hedge_maker_contract.fee_oracleUnits_to_hedge, expected_fee_oracleUnits_to_maker_at_funding)
        self.assertEqual(hedge_maker_contract.fee_oracleUnits_to_maker, expected_fee_oracleUnits_to_maker_at_funding)

        # Long Taker
        self.assertEqual(hedge_maker_contract.fee_oracleUnits_to_long, expected_fee_oracleUnits_to_taker_at_funding)
        self.assertEqual(hedge_maker_contract.fee_oracleUnits_to_taker, expected_fee_oracleUnits_to_taker_at_funding)

        # Long Maker
        self.assertEqual(long_maker_contract.fee_oracleUnits_to_long, expected_fee_oracleUnits_to_maker_at_funding)
        self.assertEqual(long_maker_contract.fee_oracleUnits_to_maker, expected_fee_oracleUnits_to_maker_at_funding)

        # Hedge Taker
        self.assertEqual(long_maker_contract.fee_oracleUnits_to_hedge, expected_fee_oracleUnits_to_taker_at_funding)
        self.assertEqual(long_maker_contract.fee_oracleUnits_to_taker, expected_fee_oracleUnits_to_taker_at_funding)


# Valid timestamp, just before maturity
BEFORE_MATURITY_TIMESTAMP = ScriptTimestamp(MATURITY_TIMESTAMP - 1)

# A valid maturity price, just somewhat over the start price so no chance of low liquidation
NAIVE_MATURITY_PRICE_USDEM2_PER_BCH = ScriptPriceInOracleUnitsPerBch(220_00)  # $220.00
BELOW_LIQUIDATING_PRICE_USDEM2_PER_BCH = ScriptPriceInOracleUnitsPerBch(DERIVED_LOW_LIQUIDATION_PRICE - 10)


class TestMaturityRedemption(unittest.TestCase):
    def test_matures_at_maturity_time(self):
        redemption = standard_contract_funding().redeem(
            price_timestamp=MATURITY_TIMESTAMP,
            price_oracleUnits_per_bch=NAIVE_MATURITY_PRICE_USDEM2_PER_BCH,
            force_maturity=False,
        )
        self.assertEqual(redemption.redemption_type, RedemptionType.MATURATION)

    def test_matures_at_maturity_time_even_if_liquidating_price(self):
        redemption = standard_contract_funding().redeem(
            price_timestamp=MATURITY_TIMESTAMP,
            price_oracleUnits_per_bch=DERIVED_LOW_LIQUIDATION_PRICE,
            force_maturity=False,
        )
        self.assertEqual(redemption.redemption_type, RedemptionType.MATURATION)

    def test_fails_to_mature_before_maturity_time(self):
        with self.assertRaises(UnredeemableError):
            standard_contract_funding().redeem(
                price_timestamp=BEFORE_MATURITY_TIMESTAMP,
                price_oracleUnits_per_bch=NAIVE_MATURITY_PRICE_USDEM2_PER_BCH,
                force_maturity=False,
            )

    def test_liquidates_before_maturity_time_at_low_liquidation_price(self):
        redemption = standard_contract_funding().redeem(
            price_timestamp=BEFORE_MATURITY_TIMESTAMP,
            price_oracleUnits_per_bch=DERIVED_LOW_LIQUIDATION_PRICE,
            force_maturity=False,
        )
        self.assertEqual(redemption.redemption_type, RedemptionType.LIQUIDATION)

    def test_fails_to_liquidate_before_maturity_time_at_one_over_low_liquidation_price(self):
        valid_liquidation_timestamp = ScriptTimestamp(MATURITY_TIMESTAMP - 1)
        one_over_liquidation_price = ScriptPriceInOracleUnitsPerBch(DERIVED_LOW_LIQUIDATION_PRICE + 1)
        with self.assertRaises(UnredeemableError):
            standard_contract_funding().redeem(
                price_timestamp=valid_liquidation_timestamp,
                price_oracleUnits_per_bch=one_over_liquidation_price,
                force_maturity=False,
            )

    def test_maturation_works_if_force(self):
        too_early_maturation_timestamp = ScriptTimestamp(MATURITY_TIMESTAMP - 1)
        redemption = standard_contract_funding().redeem(
            price_timestamp=too_early_maturation_timestamp,
            price_oracleUnits_per_bch=NAIVE_MATURITY_PRICE_USDEM2_PER_BCH,
            force_maturity=True,
        )
        self.assertEqual(redemption.redemption_type, RedemptionType.MATURATION)


###################
# Redeemed Contract Tests
###################
# Derived hedge payout sats and units
# Calculated as max(dust, HEDGE_COMPOSITE // CLAMPED_PRICE) (Note the integer division //)
DERIVED_HEDGE_PAYOUT_SATS = UtxoSats(45_454_545)  # 0.45454545... Bch
# Calculated as DERIVED_HEDGE_PAYOUT_SATS * MATURITY_PRICE / SATS_PER_BCH
DERIVED_HEDGE_PAYOUT_ORACLE_UNITS = UsdEM2Beta(9999.999900)  # 99.99999900000 USD

# Derived long payout sats and units
# Calculated as max(dust, TOTAL_INPUT - HEDGE_PAYOUT)
DERIVED_LONG_PAYOUT_SATS = UtxoSats(18_829_333)  # 0.18829... Bch
# Calculated as DERIVED_LONG_PAYOUT_SATS * MATURITY_PRICE / SATS_PER_BCH
DERIVED_LONG_PAYOUT_ORACLE_UNITS = UsdEM2Beta(4142.45326)
# Calculated as DERIVED_HEDGE_PAYOUT_ORACLE_UNITS + DERIVED_LONG_PAYOUT_ORACLE_UNITS
DERIVED_TOTAL_PAYOUT_ORACLE_UNITS = UsdEM2Beta(14142.4532)


def standard_maturation(
        funding: ContractFunding | None = None
):
    funding = funding if funding is not None else standard_contract_funding()
    return funding.redeem(
        price_timestamp=MATURITY_TIMESTAMP,
        price_oracleUnits_per_bch=NAIVE_MATURITY_PRICE_USDEM2_PER_BCH,
        force_maturity=False,
    )


def standard_liquidation(
    price_timestamp: ScriptTimestamp = MATURITY_TIMESTAMP,
    price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch = DERIVED_LOW_LIQUIDATION_PRICE,
    force_maturity: bool = False,
):
    return standard_contract_funding().redeem(
        price_timestamp=price_timestamp,
        price_oracleUnits_per_bch=price_oracleUnits_per_bch,
        force_maturity=force_maturity,
    )


class TestNewRedemptionDerivedValues(unittest.TestCase):
    ######################
    # Exact secondary values derived from contract parameters
    ######################
    def test_price_is_not_clamped_at_liquidation_price(self):
        liquidation = standard_liquidation()
        self.assertEqual(liquidation.clamped_end_price_oracleUnits_per_bch, liquidation.naive_end_price_oracleUnits_per_bch)

    def test_price_is_clamped_below_liquidation(self):
        liquidation = standard_liquidation(price_oracleUnits_per_bch=BELOW_LIQUIDATING_PRICE_USDEM2_PER_BCH)
        self.assertGreater(liquidation.clamped_end_price_oracleUnits_per_bch, liquidation.naive_end_price_oracleUnits_per_bch)

    def test_hedge_payout_sats_has_correct_value(self):
        self.assertEqual(standard_maturation().hedge_payout_sats, DERIVED_HEDGE_PAYOUT_SATS)

    def test_long_payout_sats_has_correct_value(self):
        self.assertEqual(standard_maturation().long_payout_sats, DERIVED_LONG_PAYOUT_SATS)

    ######################
    # Exact secondary unit values
    ######################
    def test_total_payout_units_has_correct_value(self):
        self.assertAlmostEqual(standard_maturation().total_payout_oracleUnits, DERIVED_TOTAL_PAYOUT_ORACLE_UNITS, delta=.001)

    def test_hedge_payout_units_has_correct_value(self):
        self.assertAlmostEqual(standard_maturation().hedge_payout_oracleUnits, DERIVED_HEDGE_PAYOUT_ORACLE_UNITS, delta=.001)

    def test_long_payout_units_has_correct_value(self):
        self.assertAlmostEqual(standard_maturation().long_payout_oracleUnits, DERIVED_LONG_PAYOUT_ORACLE_UNITS, delta=.001)

    ######################
    # Exact secondary gain values
    ######################
    def test_gain_sats(self):
        # Hedge Maker / Long Taker
        expected_gain_sats_to_hedge_maker_at_redemption = DERIVED_HEDGE_PAYOUT_SATS - DERIVED_HEDGE_INPUT_SATS + TOTAL_FEES_TO_MAKER_SATS
        expected_gain_sats_to_long_taker_at_redemption = DERIVED_LONG_PAYOUT_SATS - DERIVED_LONG_INPUT_SATS + TOTAL_FEES_TO_TAKER_SATS
        hedge_maker_contract = standard_maturation(standard_contract_funding(standard_contract_proposal(maker_side=Side.HEDGE)))

        # Hedge Maker
        self.assertEqual(hedge_maker_contract.hedge_gain_sats, expected_gain_sats_to_hedge_maker_at_redemption)
        self.assertEqual(hedge_maker_contract.maker_gain_sats, expected_gain_sats_to_hedge_maker_at_redemption)

        # Long Taker
        self.assertEqual(hedge_maker_contract.long_gain_sats, expected_gain_sats_to_long_taker_at_redemption)
        self.assertEqual(hedge_maker_contract.taker_gain_sats, expected_gain_sats_to_long_taker_at_redemption)

        # Long Maker / Hedge Taker
        expected_gain_sats_to_long_maker_at_redemption = DERIVED_LONG_PAYOUT_SATS - DERIVED_LONG_INPUT_SATS + TOTAL_FEES_TO_MAKER_SATS
        expected_gain_sats_to_hedge_taker_at_redemption = DERIVED_HEDGE_PAYOUT_SATS - DERIVED_HEDGE_INPUT_SATS + TOTAL_FEES_TO_TAKER_SATS
        long_maker_contract = standard_maturation(standard_contract_funding(standard_contract_proposal(maker_side=Side.LONG)))

        # Long Maker
        self.assertEqual(long_maker_contract.long_gain_sats, expected_gain_sats_to_long_maker_at_redemption)
        self.assertEqual(long_maker_contract.maker_gain_sats, expected_gain_sats_to_long_maker_at_redemption)

        # Hedge Taker
        self.assertEqual(long_maker_contract.hedge_gain_sats, expected_gain_sats_to_hedge_taker_at_redemption)
        self.assertEqual(long_maker_contract.taker_gain_sats, expected_gain_sats_to_hedge_taker_at_redemption)

    def test_gain_oracleUnits(self):
        # Hedge Maker / Long Taker
        expected_gain_oracleUnits_to_hedge_maker_at_redemption = UsdEM2Beta(
            (DERIVED_HEDGE_PAYOUT_SATS - DERIVED_HEDGE_INPUT_SATS + TOTAL_FEES_TO_MAKER_SATS) * NAIVE_MATURITY_PRICE_USDEM2_PER_BCH / SATS_PER_BCH
        )
        expected_gain_oracleUnits_to_long_taker_at_redemption = UsdEM2Beta(
            (DERIVED_LONG_PAYOUT_SATS - DERIVED_LONG_INPUT_SATS + TOTAL_FEES_TO_TAKER_SATS) * NAIVE_MATURITY_PRICE_USDEM2_PER_BCH / SATS_PER_BCH
        )
        hedge_maker_contract = standard_maturation(standard_contract_funding(standard_contract_proposal(maker_side=Side.HEDGE)))

        # Hedge Maker
        self.assertEqual(hedge_maker_contract.hedge_gain_oracleUnits, expected_gain_oracleUnits_to_hedge_maker_at_redemption)
        self.assertEqual(hedge_maker_contract.maker_gain_oracleUnits, expected_gain_oracleUnits_to_hedge_maker_at_redemption)

        # Long Taker
        self.assertEqual(hedge_maker_contract.long_gain_oracleUnits, expected_gain_oracleUnits_to_long_taker_at_redemption)
        self.assertEqual(hedge_maker_contract.taker_gain_oracleUnits, expected_gain_oracleUnits_to_long_taker_at_redemption)

        # Long Maker / Hedge Taker
        expected_gain_oracleUnits_to_long_maker_at_redemption = UsdEM2Beta(
            (DERIVED_LONG_PAYOUT_SATS - DERIVED_LONG_INPUT_SATS + TOTAL_FEES_TO_MAKER_SATS) * NAIVE_MATURITY_PRICE_USDEM2_PER_BCH / SATS_PER_BCH
        )
        expected_gain_oracleUnits_to_hedge_taker_at_redemption = UsdEM2Beta(
            (DERIVED_HEDGE_PAYOUT_SATS - DERIVED_HEDGE_INPUT_SATS + TOTAL_FEES_TO_TAKER_SATS) * NAIVE_MATURITY_PRICE_USDEM2_PER_BCH / SATS_PER_BCH
        )
        long_maker_contract = standard_maturation(standard_contract_funding(standard_contract_proposal(maker_side=Side.LONG)))

        # Long Maker
        self.assertEqual(long_maker_contract.long_gain_oracleUnits, expected_gain_oracleUnits_to_long_maker_at_redemption)
        self.assertEqual(long_maker_contract.maker_gain_oracleUnits, expected_gain_oracleUnits_to_long_maker_at_redemption)

        # Hedge Taker
        self.assertEqual(long_maker_contract.hedge_gain_oracleUnits, expected_gain_oracleUnits_to_hedge_taker_at_redemption)
        self.assertEqual(long_maker_contract.taker_gain_oracleUnits, expected_gain_oracleUnits_to_hedge_taker_at_redemption)

    ######################
    # Sided versions of values
    ######################
    def test_all_sided_values(self):
        c = standard_maturation()
        self.assertEqual(c.maker_payout_sats, c.hedge_payout_sats)
        self.assertEqual(c.taker_payout_sats, c.long_payout_sats)
        self.assertEqual(c.maker_payout_oracleUnits, c.hedge_payout_oracleUnits)
        self.assertEqual(c.taker_payout_oracleUnits, c.long_payout_oracleUnits)
        self.assertEqual(c.maker_gain_sats, c.hedge_gain_sats)
        self.assertEqual(c.taker_gain_sats, c.long_gain_sats)
        self.assertEqual(c.maker_gain_oracleUnits, c.hedge_gain_oracleUnits)
        self.assertEqual(c.taker_gain_oracleUnits, c.long_gain_oracleUnits)
        self.assertEqual(c.maker_gain_percent_of_own_input, c.hedge_gain_percent_of_own_input)
        self.assertEqual(c.taker_gain_percent_of_own_input, c.long_gain_percent_of_own_input)

        c = standard_maturation(standard_contract_funding(standard_contract_proposal(maker_side=Side.LONG)))
        self.assertEqual(c.maker_payout_sats, c.long_payout_sats)
        self.assertEqual(c.taker_payout_sats, c.hedge_payout_sats)
        self.assertEqual(c.maker_payout_oracleUnits, c.long_payout_oracleUnits)
        self.assertEqual(c.taker_payout_oracleUnits, c.hedge_payout_oracleUnits)
        self.assertEqual(c.maker_gain_sats, c.long_gain_sats)
        self.assertEqual(c.taker_gain_sats, c.hedge_gain_sats)
        self.assertEqual(c.maker_gain_oracleUnits, c.long_gain_oracleUnits)
        self.assertEqual(c.taker_gain_oracleUnits, c.hedge_gain_oracleUnits)
        self.assertEqual(c.maker_gain_percent_of_own_input, c.long_gain_percent_of_own_input)
        self.assertEqual(c.taker_gain_percent_of_own_input, c.hedge_gain_percent_of_own_input)


if __name__ == '__main__':
    unittest.main()



# Built-in imports
from __future__ import annotations  # allow pre-definition use of types
from dataclasses import dataclass
from enum import Enum
from math import ceil
from typing import Type

# Local imports
from . import validators
from .bch_primitives import (
    DUST,
    SATS_PER_BCH,
    SCRIPT_INT_MAX_WHEN_JAVASCRIPT,
    PublicKey,
    Sats,
    ScriptTimestamp,
    UtxoSats,
)
from .fee import (
    aggregate_fee_sats_to_role,
    FeeAgreement,
)
from .javascript import round_half_up
from .oracle import (
    oracle_pubkey_to_unit_class,
    OracleUnit,
    ScriptPriceInOracleUnitsPerBch,
)
from .role import Role


class UnredeemableError(Exception):
    pass


class Side(str, Enum):
    HEDGE = 'Hedge'
    LONG = 'Long'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.__str__()

    @property
    def other_side(self) -> Side:
        # use a lookup to ensure failure with unknown side
        return {Side.HEDGE: Side.LONG, Side.LONG: Side.HEDGE}[self]

    @classmethod
    def from_string(cls, side_string: str) -> Side:
        return {'hedge': cls.HEDGE, 'long': cls.LONG}[side_string.lower()]


class NominalOracleUnitsXSatsPerBch(int):
    def __init__(self, value):
        super().__init__()
        validators.instance(value, int)  # i.e. don't allow silent coercion
        validators.less_equal(self, SCRIPT_INT_MAX_WHEN_JAVASCRIPT)
        validators.greater_equal(self, SATS_PER_BCH)  # equivalent of 1 nominal oracle unit


class LongLeverage(float):
    min_allowed: float = 1.1
    max_allowed: float = 50.0

    def __init__(self, _):
        super().__init__()
        validators.less_equal(self, self.max_allowed * 1.00001)  # some room for floating point error
        validators.greater_equal(self, self.min_allowed * 0.99999)  # some room for floating point error


# TODO: add "forced maturation" to differentiate from normal case in records?
class RedemptionType(str, Enum):
    LIQUIDATION = 'Liquidation'
    MATURATION = 'Maturation'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.__str__()

    @classmethod
    def from_string(cls, redemption_type_string: str) -> RedemptionType:
        return {'liquidation': cls.LIQUIDATION, 'maturation': cls.MATURATION}[redemption_type_string.lower()]


@dataclass(frozen=True)
class ContractProposal:
    """Details of a proposed contract between a maker and taker. Does not include any funding oriented details such as fees."""
    # Time
    start_timestamp: ScriptTimestamp
    maturity_timestamp: ScriptTimestamp

    # Position settings
    nominal_oracleUnits_x_satsPerBch: NominalOracleUnitsXSatsPerBch

    # Start price is not an actual contract parameter, but one of start price, leverage,
    # or separated side inputs are needed in order for the contract to be fully specified.
    start_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch
    low_liquidation_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch
    oracle_public_key: PublicKey

    # Relationship between Roles and Sides
    maker_side: Side

    def __post_init__(self):
        # Note: can make validation a switch with unsafe construction if needed
        self.validate()

    ###############
    # Calculated key values used in actual contract creation and execution
    ###############
    @property
    def total_input_sats(self) -> UtxoSats:
        return UtxoSats(ceil(self.nominal_oracleUnits_x_satsPerBch / self.low_liquidation_price_oracleUnits_per_bch))

    @property
    def hedge_input_sats(self) -> UtxoSats:
        # Round gets a potentially slightly more accurate number.
        # The precision to zero-sum comes from defining long position as "everything else needed to cover liquidation".
        return UtxoSats(round_half_up(self.nominal_oracleUnits_x_satsPerBch / self.start_price_oracleUnits_per_bch))

    @property
    def long_input_sats(self) -> UtxoSats:
        return UtxoSats(self.total_input_sats - self.hedge_input_sats)

    @property
    def oracle_unit_cls(self) -> Type[OracleUnit]:
        return oracle_pubkey_to_unit_class[self.oracle_public_key]

    ###############
    # Approximate unit versions of the inputs
    ###############
    @property
    def total_input_oracleUnits(self) -> OracleUnit:
        return self.oracle_unit_cls(self.effective_nominal_oracleUnits)

    @property
    def hedge_input_oracleUnits(self) -> OracleUnit:
        return self.oracle_unit_cls(self.nominal_oracleUnits_x_satsPerBch / SATS_PER_BCH)

    @property
    def long_input_oracleUnits(self) -> OracleUnit:
        return self.oracle_unit_cls(self.long_input_sats.bch * float(self.start_price_oracleUnits_per_bch))

    ###############
    # Reverse calculated effective values from contract perspective
    ###############
    @property
    def duration_seconds(self) -> int:
        return self.maturity_timestamp - self.start_timestamp

    @property
    def effective_nominal_oracleUnits(self) -> OracleUnit:
        return self.oracle_unit_cls(self.nominal_oracleUnits_x_satsPerBch / SATS_PER_BCH)

    @property
    def effective_long_leverage(self) -> LongLeverage:
        return LongLeverage(1 / (1 - self.low_liquidation_price_oracleUnits_per_bch / self.start_price_oracleUnits_per_bch))

    ###############
    # Sided views
    ###############
    @property
    def taker_side(self) -> Side:
        return self.maker_side.other_side

    @property
    def maker_input_sats(self) -> Sats:
        if self.maker_side == Side.HEDGE:
            return self.hedge_input_sats
        return self.long_input_sats

    @property
    def taker_input_sats(self) -> Sats:
        if self.taker_side == Side.HEDGE:
            return self.hedge_input_sats
        return self.long_input_sats

    @property
    def maker_input_oracleUnits(self) -> OracleUnit:
        if self.maker_side == Side.HEDGE:
            return self.hedge_input_oracleUnits
        return self.long_input_oracleUnits

    @property
    def taker_input_oracleUnits(self) -> OracleUnit:
        if self.taker_side == Side.HEDGE:
            return self.hedge_input_oracleUnits
        return self.long_input_oracleUnits

    ###############
    # Constructors
    ###############
    @staticmethod
    def new_from_intent(start_timestamp: ScriptTimestamp,
                        maturity_timestamp: ScriptTimestamp,
                        nominal_oracleUnits: OracleUnit,
                        long_leverage: LongLeverage,
                        start_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch,
                        maker_side: Side,
                        ) -> ContractProposal:
        nominal_oracleUnits_x_satsPerBch = NominalOracleUnitsXSatsPerBch(round_half_up(nominal_oracleUnits * SATS_PER_BCH))
        low_liquidation_price_oracleUnits_per_bch = ScriptPriceInOracleUnitsPerBch(round_half_up(float(start_price_oracleUnits_per_bch) * (1 - 1 / long_leverage)))
        return ContractProposal(
            start_timestamp=start_timestamp,
            maturity_timestamp=maturity_timestamp,
            nominal_oracleUnits_x_satsPerBch=nominal_oracleUnits_x_satsPerBch,
            start_price_oracleUnits_per_bch=start_price_oracleUnits_per_bch,
            low_liquidation_price_oracleUnits_per_bch=low_liquidation_price_oracleUnits_per_bch,
            oracle_public_key=nominal_oracleUnits.public_key,
            maker_side=maker_side,
        )

    def fund(self, fee_agreements: list[FeeAgreement]) -> ContractFunding:
        return ContractFunding(
            base_proposal=self,
            fee_agreements=tuple(fee_agreements),
        )

    def validate(self):
        # Timing
        min_contract_duration_seconds = 60
        if not (self.duration_seconds >= min_contract_duration_seconds):
            raise ValueError(f'contract duration is {self.duration_seconds} s but it must be >= {min_contract_duration_seconds} s')


@dataclass(frozen=True)
class ContractFunding:
    """Funding details and actions, typically derived from a contract proposal."""
    base_proposal: ContractProposal
    fee_agreements: tuple[FeeAgreement, ...]

    @property
    def fee_sats_to_maker(self):
        return aggregate_fee_sats_to_role(self.fee_agreements, Role.MAKER)

    @property
    def fee_sats_to_taker(self):
        return aggregate_fee_sats_to_role(self.fee_agreements, Role.TAKER)

    @property
    def fee_sats_to_hedge(self):
        if self.base_proposal.maker_side == Side.HEDGE:
            return self.fee_sats_to_maker
        return self.fee_sats_to_taker

    @property
    def fee_sats_to_long(self):
        if self.base_proposal.maker_side == Side.LONG:
            return self.fee_sats_to_maker
        return self.fee_sats_to_taker

    ###############
    # Unit value calculations
    ###############
    @property
    def fee_oracleUnits_to_maker(self):
        fee_bch = aggregate_fee_sats_to_role(self.fee_agreements, Role.MAKER).bch
        fee_oracleUnits = self.base_proposal.oracle_unit_cls(fee_bch * float(self.base_proposal.start_price_oracleUnits_per_bch))
        return fee_oracleUnits

    @property
    def fee_oracleUnits_to_taker(self):
        fee_bch = aggregate_fee_sats_to_role(self.fee_agreements, Role.TAKER).bch
        fee_oracleUnits = self.base_proposal.oracle_unit_cls(fee_bch * float(self.base_proposal.start_price_oracleUnits_per_bch))
        return fee_oracleUnits

    @property
    def fee_oracleUnits_to_hedge(self):
        if self.base_proposal.maker_side == Side.HEDGE:
            return self.fee_oracleUnits_to_maker
        return self.fee_oracleUnits_to_taker

    @property
    def fee_oracleUnits_to_long(self):
        if self.base_proposal.maker_side == Side.LONG:
            return self.fee_oracleUnits_to_maker
        return self.fee_oracleUnits_to_taker

    ###############
    # Specialized calculations
    ###############
    def effective_adjusted_price_oracleUnits_per_bch(self, role: Role) -> float:
        # The direction of price adjustment depends on which side is maker/taker.
        # Without context, positive fees will cause the formula to push the price up.
        # However, in context, the higher initial effective price benefits a hedge
        # maker specifically. Therefore, when the maker is long, we need to reverse
        # the direction of the fees so that the benefit always goes to the maker.
        #
        # Unfortunately, the convention of most of the code is "fees to"
        # instead of "fees from" which the original equation was created with.
        # Therefore, there is an extra -1 causing directionality to look reversed:
        direction = 1
        if self.base_proposal.maker_side == Side.HEDGE:
            direction = -1

        # The calculation is not especially intuitive. Here is the breakdown:
        # Naming:
        # nominal units = u
        # fees paid in bch = fP = Σ(fP1...fPm)
        # fees received in bch = fR = Σ(fR1...fRn)
        # original price = pS
        # adjusted price = pA
        #
        # (fP - fR) = u (1/pS - 1/pA)
        # (fP - fR)/u = 1/pS - 1/pA
        # 1/pA = 1/pS - (fP - fR)/u
        # 1/pA = (u - pS(fP - fR)) / (u pS)
        # pA = u pS / (u - pS(fP - fR))

        u = self.base_proposal.effective_nominal_oracleUnits
        pS = float(self.base_proposal.start_price_oracleUnits_per_bch)
        total_fees_to_role_bch = direction * aggregate_fee_sats_to_role(self.fee_agreements, role).bch
        adjusted_price_oracleUnits_per_bch = u * pS / (u - pS * total_fees_to_role_bch)
        return adjusted_price_oracleUnits_per_bch

    ###############
    # Actions
    ###############
    def redeem(self,
               price_timestamp: ScriptTimestamp,
               price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch,
               force_maturity: bool,
               ) -> ContractRedemption:
        """Redeem the contract according to market conditions or raise an unredeemable error for invalid conditions."""
        reached_maturity_time = price_timestamp >= self.base_proposal.maturity_timestamp
        reached_liquidation_price = price_oracleUnits_per_bch <= self.base_proposal.low_liquidation_price_oracleUnits_per_bch
        if reached_maturity_time or force_maturity:
            # Maturation, even in the case of a liquidation price
            return ContractRedemption(
                base_funding=self,
                end_price_timestamp=price_timestamp,
                naive_end_price_oracleUnits_per_bch=price_oracleUnits_per_bch,
                redemption_type=RedemptionType.MATURATION,
            )
        elif reached_liquidation_price:
            # Liquidation
            return ContractRedemption(
                base_funding=self,
                end_price_timestamp=price_timestamp,
                naive_end_price_oracleUnits_per_bch=price_oracleUnits_per_bch,
                redemption_type=RedemptionType.LIQUIDATION,
            )

        # If conditions did not allow for redemption, throw an error
        raise UnredeemableError


@dataclass(frozen=True)
class ContractRedemption:
    """Outcome of a redeemed contract, especially with respect to the two counterparties."""
    base_funding: ContractFunding
    end_price_timestamp: ScriptTimestamp
    naive_end_price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch
    redemption_type: RedemptionType

    @property
    def clamped_end_price_oracleUnits_per_bch(self) -> ScriptPriceInOracleUnitsPerBch:
        return max(self.naive_end_price_oracleUnits_per_bch, self.base_funding.base_proposal.low_liquidation_price_oracleUnits_per_bch)

    @property
    def total_payout_sats(self) -> UtxoSats:
        # Note: This can be different from total input in the case of liquidation where dust protection is pulled in from outside the contract
        return UtxoSats(self.hedge_payout_sats + self.long_payout_sats)

    @property
    def hedge_payout_sats(self) -> UtxoSats:
        # Note that due to dust safety in the contract, the total actual payout can be greater than total inputs.
        # In reality, the extra dust is covered by an amount sitting on the contract that the contract is not aware of.
        # Use divmod (instead of //) to make it crystal clear this represents integer division of the contract.
        unsafe_hedge_payout_sats, _ = divmod(self.base_funding.base_proposal.nominal_oracleUnits_x_satsPerBch, self.clamped_end_price_oracleUnits_per_bch)
        safe_hedge_payout_sats = max(DUST, unsafe_hedge_payout_sats)
        return UtxoSats(safe_hedge_payout_sats)

    @property
    def long_payout_sats(self) -> UtxoSats:
        # Note that due to dust safety in the contract, the total actual payout can be greater than total inputs.
        # In reality, the extra dust is covered by an amount sitting on the contract that the contract is not aware of.
        unsafe_long_payout_sats = self.base_funding.base_proposal.total_input_sats - self.hedge_payout_sats
        safe_long_payout_sats = max(DUST, unsafe_long_payout_sats)
        return UtxoSats(safe_long_payout_sats)

    ###############
    # Approximate unit versions of the outputs
    ###############
    @property
    def total_payout_oracleUnits(self) -> OracleUnit:
        return self.base_funding.base_proposal.oracle_unit_cls(self.total_payout_sats.bch * float(self.clamped_end_price_oracleUnits_per_bch))

    @property
    def hedge_payout_oracleUnits(self) -> OracleUnit:
        return self.base_funding.base_proposal.oracle_unit_cls(self.hedge_payout_sats.bch * float(self.clamped_end_price_oracleUnits_per_bch))

    @property
    def long_payout_oracleUnits(self) -> OracleUnit:
        return self.base_funding.base_proposal.oracle_unit_cls(self.long_payout_sats.bch * float(self.clamped_end_price_oracleUnits_per_bch))

    ###############
    # Gains - see funding class for details of fee and gain calculations
    ###############
    @property
    def hedge_gain_sats(self) -> Sats:
        payout_sats = self.hedge_payout_sats
        input_sats = self.base_funding.base_proposal.hedge_input_sats
        fee_sats = self.base_funding.fee_sats_to_hedge
        return Sats(payout_sats - input_sats + fee_sats)

    @property
    def long_gain_sats(self) -> Sats:
        payout_sats = self.long_payout_sats
        input_sats = self.base_funding.base_proposal.long_input_sats
        fee_sats = self.base_funding.fee_sats_to_long
        return Sats(payout_sats - input_sats + fee_sats)

    ###############
    # Approximate unit versions of the gains
    ###############
    @property
    def hedge_gain_oracleUnits(self) -> OracleUnit:
        # Note that we use naive end price. This represents reality of slippage in liquidations.
        return self.base_funding.base_proposal.oracle_unit_cls(self.hedge_gain_sats.bch * float(self.naive_end_price_oracleUnits_per_bch))

    @property
    def long_gain_oracleUnits(self) -> OracleUnit:
        # Note that we use naive end price. This represents reality of slippage in liquidations.
        return self.base_funding.base_proposal.oracle_unit_cls(self.long_gain_sats.bch * float(self.naive_end_price_oracleUnits_per_bch))

    ###############
    # Relative gains
    ###############
    @property
    def hedge_gain_percent_of_own_input(self) -> float:
        return 100.0 * float(self.hedge_gain_sats) / float(self.base_funding.base_proposal.hedge_input_sats)

    @property
    def long_gain_percent_of_own_input(self) -> float:
        return 100.0 * float(self.long_gain_sats) / float(self.base_funding.base_proposal.long_input_sats)

    ###############
    # Sided views on all of the above
    ###############
    @property
    def maker_payout_sats(self) -> UtxoSats:
        if self.base_funding.base_proposal.maker_side == Side.HEDGE:
            return self.hedge_payout_sats
        return self.long_payout_sats

    @property
    def taker_payout_sats(self) -> UtxoSats:
        if self.base_funding.base_proposal.taker_side == Side.HEDGE:
            return self.hedge_payout_sats
        return self.long_payout_sats

    @property
    def maker_payout_oracleUnits(self) -> OracleUnit:
        if self.base_funding.base_proposal.maker_side == Side.HEDGE:
            return self.hedge_payout_oracleUnits
        return self.long_payout_oracleUnits

    @property
    def taker_payout_oracleUnits(self) -> OracleUnit:
        if self.base_funding.base_proposal.taker_side == Side.HEDGE:
            return self.hedge_payout_oracleUnits
        return self.long_payout_oracleUnits

    @property
    def maker_gain_sats(self) -> Sats:
        if self.base_funding.base_proposal.maker_side == Side.HEDGE:
            return self.hedge_gain_sats
        return self.long_gain_sats

    @property
    def taker_gain_sats(self) -> Sats:
        if self.base_funding.base_proposal.taker_side == Side.HEDGE:
            return self.hedge_gain_sats
        return self.long_gain_sats

    @property
    def maker_gain_oracleUnits(self) -> OracleUnit:
        if self.base_funding.base_proposal.maker_side == Side.HEDGE:
            return self.hedge_gain_oracleUnits
        return self.long_gain_oracleUnits

    @property
    def taker_gain_oracleUnits(self) -> OracleUnit:
        if self.base_funding.base_proposal.taker_side == Side.HEDGE:
            return self.hedge_gain_oracleUnits
        return self.long_gain_oracleUnits

    @property
    def maker_gain_percent_of_own_input(self) -> float:
        if self.base_funding.base_proposal.maker_side == Side.HEDGE:
            return self.hedge_gain_percent_of_own_input
        return self.long_gain_percent_of_own_input

    @property
    def taker_gain_percent_of_own_input(self) -> float:
        if self.base_funding.base_proposal.taker_side == Side.HEDGE:
            return self.hedge_gain_percent_of_own_input
        return self.long_gain_percent_of_own_input



from unittest import TestCase as _TestCaseForAssertions


_a = _TestCaseForAssertions()

instance = _a.assertIsInstance
equal = _a.assertEqual
almost_equal = _a.assertAlmostEqual
less_equal = _a.assertLessEqual
greater_equal = _a.assertGreaterEqual



# Built-in imports
from enum import Enum


class Role(str, Enum):
    MAKER = 'Maker'
    TAKER = 'Taker'
    SETTLEMENT_SERVICE = 'Settlement Service'
    # OTHER = 'Other'

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.__str__()



# Built-in imports
from dataclasses import dataclass
from types import NoneType
from typing import Sequence

# Local imports
from .bch_primitives import Sats
from .role import Role

# TODO: make an aggregate "FeeAgreements" that packs them up and makes them available in parts if desired
#       or in total with full consistency between parts and total. i.e. no discrepancy between total and sum of parts


@dataclass(frozen=True)
class FeeRequirement:
    name: str
    amount_sats: Sats
    # note that amount can be negative indicating the "to" role is paying, not receiving
    receiving: Role

    def __str__(self):
        return f'FeeRequirement ({self.name}): ____ --> {self.amount_sats} Sats ({self.amount_sats.bch} BCH) --> {self.receiving}'

    def __repr__(self):
        return self.__str__()

    def make_agreement(self, paying: Role):
        return FeeAgreement(
            name=self.name,
            amount_sats=self.amount_sats,
            receiving=self.receiving,
            paying=paying,
        )


@dataclass(frozen=True)
class FeeAgreement(FeeRequirement):
    paying: Role

    def __str__(self):
        return f'FeeAgreement ({self.name}): {self.paying} --> {self.amount_sats} Sats ({self.amount_sats.bch} BCH) --> {self.receiving}'

    def __repr__(self):
        return self.__str__()


def aggregate_fee_sats_to_role(
        fees: Sequence[FeeAgreement],
        role: Role,
        fee_name: str | NoneType = None,
) -> Sats:
    sats_to_role = sum(fee.amount_sats for fee in fees_to_role(fees, role, fee_name))
    sats_from_role = sum(fee.amount_sats for fee in fees_from_role(fees, role, fee_name))
    return Sats(sats_to_role - sats_from_role)


def fees_to_role(
        fees: Sequence[FeeAgreement],
        role: Role,
        fee_name: str | None = None) -> list[FeeAgreement]:
    return [
        fee for fee in fees
        if (fee.receiving == role) and ((fee_name is None) or fee.name == fee_name)
    ]


def fees_from_role(
        fees: Sequence[FeeAgreement],
        role: Role,
        fee_name: str | None = None) -> list[FeeAgreement]:
    return [
        fee for fee in fees
        if (fee.paying == role) and ((fee_name is None) or fee.name == fee_name)
    ]



# Built-in imports
import unittest

# Local imports
from .bch_primitives import (
    PublicKey,
    ScriptTimestamp,
)
from .oracle import (
    Pricepoint,
    ScriptPriceInOracleUnitsPerBch,
)


# Known price message for Usd(e-2) oracle:
#   public key: 02d3c1de9d4bc77d6c3608cbe44d10138c7488e592dc2b1e10a6cf0e92c2ecb047
#   signature: 9f66be2e0b542325303102afb584cdcee591b84a9ec6aba311e73b0f1cbba8d5d71c701b7b6c3230e4814170d79cafea0d90ea0e78d666d8f50dd80b28caf00e
#   message: 2f2635633fa80300a4a50300242d0000
#       timestamp:        1664427567
#       message sequence: 239679
#       price sequence:   239012
#       price:            11556
known_pricepoint = Pricepoint(
    oracle_pubkey=PublicKey('02d3c1de9d4bc77d6c3608cbe44d10138c7488e592dc2b1e10a6cf0e92c2ecb047'),
    message='2f2635633fa80300a4a50300242d0000',
    signature='9f66be2e0b542325303102afb584cdcee591b84a9ec6aba311e73b0f1cbba8d5d71c701b7b6c3230e4814170d79cafea0d90ea0e78d666d8f50dd80b28caf00e',
)


class TestPricepoint(unittest.TestCase):
    def test_new_from_details_has_all_correct_attributes_and_properties(self):
        input_oracle_pubkey = PublicKey('02d3c1de9d4bc77d6c3608cbe44d10138c7488e592dc2b1e10a6cf0e92c2ecb047')
        input_timestamp = ScriptTimestamp(1664427567)
        input_price_oracleUnits_per_bch = ScriptPriceInOracleUnitsPerBch(11556)
        input_message_sequence = 239679
        input_price_sequence = 239012
        input_signature = '9f66be2e0b542325303102afb584cdcee591b84a9ec6aba311e73b0f1cbba8d5d71c701b7b6c3230e4814170d79cafea0d90ea0e78d666d8f50dd80b28caf00e'

        expected_oracleUnit_name = 'USD(e-2)'
        expected_message = '2f2635633fa80300a4a50300242d0000'
        expected_standardUnits_per_bch = 115.56

        p = Pricepoint.new_from_details(
            oracle_pubkey=input_oracle_pubkey,
            timestamp=input_timestamp,
            price_oracleUnits_per_bch=input_price_oracleUnits_per_bch,
            message_sequence=input_message_sequence,
            price_sequence=input_price_sequence,
            signature=input_signature,
        )

        # confirm all the primary attributes and properties
        self.assertEqual(p.oracle_pubkey, input_oracle_pubkey)
        self.assertEqual(p.message, expected_message)
        self.assertEqual(p.signature, input_signature)
        self.assertEqual(p.oracleUnit_name, expected_oracleUnit_name)
        self.assertEqual(p.timestamp, input_timestamp)
        self.assertEqual(p.message_sequence, input_message_sequence)
        self.assertEqual(p.price_sequence, input_price_sequence)
        self.assertEqual(p.price_oracleUnits_per_bch, input_price_oracleUnits_per_bch)
        self.assertEqual(p.price_standardUnits_per_bch, expected_standardUnits_per_bch)


if __name__ == '__main__':
    unittest.main()



# Built-in imports
from __future__ import annotations
from dataclasses import dataclass
from functools import cached_property
from typing import (
    Sequence,
    Type,
)

# Library imports
from arrow import Arrow

# Local imports
from . import validators
from .bch_primitives import (
    PublicKey,
    ScriptTimestamp,
)

# Oracle prices are stored and limited to a 4 byte field.
# It may be unsigned, but I'm not sure so I use 31 bits to be safe
ORACLE_MAX_PRICE = (2 ** 31) - 1
GENERIC_4_BYTE_NUMBER = 2 ** 30


@dataclass(frozen=False)
class StrictPricepoints:
    """Sorted, unique pricepoints all from the same oracle"""
    all: Sequence[Pricepoint, ...]

    def __post_init__(self):
        # first validate oracle pubkeys (definitely die if mixed oracles)
        if self.all:
            oracle_pubkey = self.all[0].oracle_pubkey
            for p in self.all:
                validators.equal(p.oracle_pubkey, oracle_pubkey)

        # enforce uniqueness by timestamp
        lookup = {}
        for p in self.all:
            existing = lookup.get(p.timestamp)
            if existing is None:
                lookup[p.timestamp] = p
            else:
                if p == existing:
                    pass  # it's just a duplicate
                else:
                    raise ValueError(f'same timestamp, same oracle, different value: {p} vs {existing}')

        # enforce sorting
        self.all = tuple(sorted(lookup.values(), key=lambda p: p.timestamp))

    def __str__(self):
        return f'StrictPricepoints: {len(self.all)} prices for {self.all[0].oracleUnit_name}'

    def __repr__(self):
        return self.__str__()


def interleave_pricepoints(pricepoints_by_oracle_pubkey: dict[PublicKey, StrictPricepoints]) -> list[Pricepoint]:
    # TODO: There could be duplicates or bad overlapping messages hiding in this
    #       To solve that, make a formalized PricepointsByOracle class that strongly validates match between pubkey and pricepoints
    # let the optimized sorting algorithm sort it out even though we could do it in O(n) manually
    all_pricepoints = sum((list(p.all) for p in pricepoints_by_oracle_pubkey.values()), start=[])
    all_pricepoints.sort(key=lambda p: p.timestamp)
    return all_pricepoints


@dataclass(frozen=True)
class Pricepoint:
    oracle_pubkey: PublicKey
    message: str
    signature: str | None

    def __post_init__(self):
        # cheat for some early validation
        _, _, _, _ = self._message_parts

    def __str__(self):
        return f'{self.price_oracleUnits_per_bch} {self.oracleUnit_name} @ {self.timestamp} ({Arrow.utcfromtimestamp(self.timestamp)})'

    def __repr__(self):
        return self.__str__()

    @property
    def oracleUnit_name(self) -> str:
        return oracle_pubkey_to_unit_class[self.oracle_pubkey].name_of_oracleUnits

    @property
    def standardUnit_name(self) -> str:
        return oracle_pubkey_to_unit_class[self.oracle_pubkey].name_of_standardUnits

    @property
    def timestamp(self) -> ScriptTimestamp:
        return self._message_parts[0]

    @property
    def message_sequence(self) -> int:
        return self._message_parts[1]

    @property
    def price_sequence(self) -> int:
        return self._message_parts[2]

    @property
    def price_oracleUnits_per_bch(self) -> ScriptPriceInOracleUnitsPerBch:
        return ScriptPriceInOracleUnitsPerBch(self._message_parts[3])

    @property
    def price_standardUnits_per_bch(self) -> float:
        return float(self.price_oracleUnits_per_bch) / oracle_pubkey_to_unit_class[self.oracle_pubkey].oracleUnits_per_standardUnit

    @cached_property
    def _message_parts(self):
        # timestamp, message_sequence, price_sequence, price
        return parse_oracle_price_message(self.message)

    @classmethod
    def new_from_details(cls,
                         oracle_pubkey: PublicKey,
                         timestamp: ScriptTimestamp,
                         price_oracleUnits_per_bch: ScriptPriceInOracleUnitsPerBch,
                         message_sequence: int  = GENERIC_4_BYTE_NUMBER,
                         price_sequence: int  = GENERIC_4_BYTE_NUMBER,
                         signature: str | None = None,
                         ) -> Pricepoint:
        message_parts = []
        for data in (timestamp, message_sequence, price_sequence, price_oracleUnits_per_bch):
            message_parts.append(_swap_hex_endianness(f'{data:08x}'))
        message = ''.join(message_parts)
        pricepoint = Pricepoint(oracle_pubkey=oracle_pubkey, message=message, signature=signature)

        # do a sanity reverse check on the parsed values
        validators.equal(pricepoint.timestamp, timestamp)
        validators.equal(pricepoint.price_oracleUnits_per_bch, price_oracleUnits_per_bch)
        validators.equal(pricepoint.message_sequence, message_sequence)
        validators.equal(pricepoint.price_sequence, price_sequence)

        return pricepoint


class OracleUnit(float):
    name_of_oracleUnits: str = None
    name_of_standardUnits: str = None
    oracleUnits_per_standardUnit: int = None
    public_key: PublicKey = None

    def __init__(self, value):
        super().__init__()
        validators.instance(value, float)  # i.e. don't allow silent coercion
        # validators.greater_equal(self, 0)

    def __eq__(self, other: OracleUnit):
        if self.name_of_oracleUnits != other.name_of_oracleUnits:
            return False
        if self.name_of_standardUnits != other.name_of_standardUnits:
            return False
        if self.oracleUnits_per_standardUnit != other.oracleUnits_per_standardUnit:
            return False
        if self.public_key != other.public_key:
            return False
        if float(self) != float(other):
            return False
        return True

    def __hash__(self):
        return self.public_key

    @property
    def in_standard_units(self) -> float:
        return float(self) / self.oracleUnits_per_standardUnit


class ScriptPriceInOracleUnitsPerBch(int):
    def __new__(cls, value, *args, **kwargs):
        return super().__new__(cls, value)

    def __init__(self, value):
        super().__init__()
        validators.instance(value, int)  # i.e. don't allow silent coercion
        validators.less_equal(self, ORACLE_MAX_PRICE)
        validators.greater_equal(self, 1)


class BtcEM6Beta(OracleUnit):
    name_of_oracleUnits = 'BTC(e-6)'
    name_of_standardUnits = 'BTC'
    oracleUnits_per_standardUnit = 1000000
    public_key = PublicKey('03c22127c967bb28ec518fcc225164100df40470a1f6b457cd3a85adb051dcaa56')


class CnyEM2Beta(OracleUnit):
    name_of_oracleUnits = 'CNY(e-2)'
    name_of_standardUnits = 'CNY'
    oracleUnits_per_standardUnit = 100
    public_key = PublicKey('03c5e3e6a2fe9ed9be3c71a11e7808cf8428bc9ca48808d05a6fa2526865964f06')


class DogeEM1Beta(OracleUnit):
    name_of_oracleUnits = 'DOGE(e-1)'
    name_of_standardUnits = 'DOGE'
    oracleUnits_per_standardUnit = 10
    public_key = PublicKey('0330779426934d4fe5d18a3721e9eae246150501ebc537e866d2841369daeb0691')


class EthEM5Beta(OracleUnit):
    name_of_oracleUnits = 'ETH(e-5)'
    name_of_standardUnits = 'ETH'
    oracleUnits_per_standardUnit = 100000
    public_key = PublicKey('03518d199a4ca5dc06ecb1068416acde321df1b8d6f09149744b1e0fb38c92c92c')


class InrEM0Beta(OracleUnit):
    name_of_oracleUnits = 'INR'
    name_of_standardUnits = 'INR'
    oracleUnits_per_standardUnit = 1
    public_key = PublicKey('03994dc2c759375e98afbf5049383cd987001c346d0f11aa262c105874fb1390c3')


class UsdEM2Beta(OracleUnit):
    name_of_oracleUnits = 'USD(e-2)'
    name_of_standardUnits = 'USD'
    oracleUnits_per_standardUnit = 100
    public_key = PublicKey('02d3c1de9d4bc77d6c3608cbe44d10138c7488e592dc2b1e10a6cf0e92c2ecb047')


class UsdEM2Testing(OracleUnit):
    name_of_oracleUnits = 'USD(e-2)'
    name_of_standardUnits = 'USD'
    oracleUnits_per_standardUnit = 100
    public_key = PublicKey('03e4a5b4a1d9365492dfcad9a05adcdea259acb51a20eabb84584f7aa3c7009ebe')


class XagEM3Beta(OracleUnit):
    name_of_oracleUnits = 'XAG(e-3)'
    name_of_standardUnits = 'XAG'
    oracleUnits_per_standardUnit = 1000
    public_key = PublicKey('03e9342b4d07dc35db0f555b80e19645b2a2a95a22675b50ead248d551a900fdec')


class XauEM5Beta(OracleUnit):
    name_of_oracleUnits = 'XAU(e-5)'
    name_of_standardUnits = 'XAU'
    oracleUnits_per_standardUnit = 100000
    public_key = PublicKey('03e980928f14fc98e1f9d75d15f0b67dc58cdd3f5c641b8f825b146bcc04bd232c')


known_unit_classes = [BtcEM6Beta, CnyEM2Beta, DogeEM1Beta, EthEM5Beta, InrEM0Beta, UsdEM2Beta, UsdEM2Testing, XagEM3Beta, XauEM5Beta]
oracle_pubkey_to_unit_class: dict[PublicKey, Type[OracleUnit]] = {
    u.public_key: u
    for u in known_unit_classes
}


def parse_oracle_price_message(msg_hex: str) -> tuple[ScriptTimestamp, int, int, ScriptPriceInOracleUnitsPerBch]:
    """
    [0, 8) timestamp (4 bytes)
    [8, 16) message sequence (4 bytes)
    [16, 24) price sequence (4 bytes)
    [24, 32) price (4 bytes)
    """
    if len(msg_hex) == 32:
        # might be a price message
        pass
    else:
        print(f'unknown oracle message. probably meta: {msg_hex}')

    timestamp = ScriptTimestamp(int(_swap_hex_endianness(msg_hex[0:8]), 16))
    msg_sequence = int(_swap_hex_endianness(msg_hex[8:16]), 16)
    price_sequence = int(_swap_hex_endianness(msg_hex[16:24]), 16)
    price = ScriptPriceInOracleUnitsPerBch(int(_swap_hex_endianness(msg_hex[24:]), 16))
    return timestamp, msg_sequence, price_sequence, price


def _swap_hex_endianness(data_hex: str) -> str:
    data_bytes = bytearray.fromhex(data_hex)
    data_bytes.reverse()
    data_hex_reversed = ''.join(format(x, '02x') for x in data_bytes)
    return data_hex_reversed
